# -*- coding: utf-8 -*-

import logging
from threading import Event, Thread


class StoppableThread(Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, **kwargs):
        self._stop_event = Event()

        if "logger" in kwargs:
            self.logger = kwargs["logger"]
        elif "logger_name" in kwargs:
            self.logger = logging.getLogger(kwargs["logger_name"])
        else:
            self.logger = logging.getLogger(__name__)

        filtered_kwargs = {
            k: v for k, v in kwargs.items() if k not in ("logger", "logger_name")
        }

        Thread.__init__(self, **filtered_kwargs)

    def stop(self):
        """Sets the stop event"""
        self.logger.debug("Stopping thread: %s", self.name)
        self._stop_event.set()

    def stopped(self):
        """Determines if stop has been called yet."""
        return self._stop_event.isSet()

    def wait(self, timeout=None):
        """Delegate wait call to threading.Event"""
        return self._stop_event.wait(timeout)



# -*- coding: utf-8 -*-
import json
import logging
from argparse import ArgumentParser

from yapconf import YapconfSpec
from yapconf.exceptions import YapconfItemNotFound

from brewtils.errors import ValidationError, _deprecate
from brewtils.rest import normalize_url_prefix
from brewtils.specification import SPECIFICATION, _CONNECTION_SPEC

logger = logging.getLogger(__name__)


def get_argument_parser():
    """Get an ArgumentParser pre-populated with Brewtils arguments

    This is helpful if you're expecting additional command line arguments to
    a plugin startup script.

    This enables doing something like::

        def main():
            parser = get_argument_parser()
            parser.add_argument('positional_arg')

            parsed_args = parser.parse_args(sys.argv[1:])

            # Now you can use the extra argument
            client = MyClient(parsed_args.positional_arg)

            # But you'll need to be careful when using the 'normal' Brewtils
            # configuration loading methods:

            # Option 1: Tell Brewtils about your customized parser
            connection = get_connection_info(cli_args=sys.argv[1:],
                                             argument_parser=parser)

            # Option 2: Use the parsed CLI as a dictionary
            connection = get_connection_info(**vars(parsed_args))

            # Now specify connection kwargs like normal
            plugin = RemotePlugin(client, name=...
                                  **connection)

    IMPORTANT: Note that in both cases the returned ``connection`` object **will
    not** contain your new value. Both options just prevent normal CLI parsing
    from failing on the unknown argument.

    Returns:
        ArgumentParser: Argument parser with Brewtils arguments loaded
    """
    parser = ArgumentParser()

    YapconfSpec(SPECIFICATION).add_arguments(parser)

    return parser


def get_connection_info(cli_args=None, argument_parser=None, **kwargs):
    """Wrapper around ``load_config`` that returns only connection parameters

    Args:
        cli_args (list, optional): List of command line arguments for
            configuration loading
        argument_parser (ArgumentParser, optional): Argument parser to use when
            parsing cli_args. Supplying this allows adding additional arguments
            prior to loading the configuration. This can be useful if your
            startup script takes additional arguments.
        **kwargs: Additional configuration overrides

    Returns:
        dict: Parameters needed to make a connection to Beergarden
    """
    config = load_config(cli_args=cli_args, argument_parser=argument_parser, **kwargs)

    return {key: config[key] for key in _CONNECTION_SPEC}


def load_config(
    cli_args=True, environment=True, argument_parser=None, bootstrap=False, **kwargs
):
    """Load configuration using Yapconf

    Configuration will be loaded from these sources, with earlier sources having
    higher priority:

        1. ``**kwargs`` passed to this method
        2. Command line arguments (if ``cli_args`` argument is not False)
        3. Environment variables using the ``BG_`` prefix (if ``environment`` argument
            is not False)
        4. Default values in the brewtils specification

    Args:
        cli_args (Union[bool, list], optional): Specifies whether command line should be
            used as a configuration source
            - True: Argparse will use the standard sys.argv[1:]
            - False: Command line arguments will be ignored when loading configuration
            - List of strings: Will be parsed as CLI args (instead of using sys.argv)
        environment (bool): Specifies whether environment variables (with the ``BG_``
            prefix) should be used when loading configuration
        argument_parser (ArgumentParser, optional, deprecated): Argument parser to use
            when parsing cli_args. Supplying this allows adding additional arguments
            prior to loading the configuration. This can be useful if your
            startup script takes additional arguments. See get_argument_parser
            for additional information.
        **kwargs: Additional configuration overrides

    Returns:
        box.Box: The resolved configuration object
    """
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")

    sources = []

    if kwargs:
        # First deprecate / translate items with multiple names
        mangled_kwargs = _translate_kwargs(**kwargs)

        # Metadata is a little weird because yapconf doesn't support raw dicts, so we
        # need to make it a json string in that case
        metadata = kwargs.get("metadata")
        if isinstance(metadata, dict):
            mangled_kwargs["metadata"] = json.dumps(metadata)

        sources.append(("kwargs", mangled_kwargs))

    if cli_args:
        if cli_args is True:
            sources.append("CLI")
        else:
            if not argument_parser:
                argument_parser = ArgumentParser()
                spec.add_arguments(argument_parser)

            parsed_args, unknown = argument_parser.parse_known_args(cli_args)
            sources.append(("cli_args", vars(parsed_args)))

    if environment:
        sources.append("ENVIRONMENT")

    try:
        config = spec.load_config(*sources, bootstrap=bootstrap)
    except YapconfItemNotFound as ex:
        if ex.item.name == "bg_host":
            raise ValidationError(
                "Unable to create a plugin without a Beer-garden host. Please specify "
                "one on the command line (--bg-host), in the environment (BG_HOST), or "
                "in kwargs (bg_host)."
            )
        raise

    # Make sure the url_prefix is normal
    if "bg_url_prefix" in config:
        config.bg_url_prefix = normalize_url_prefix(config.bg_url_prefix)

    return config


def _translate_kwargs(**kwargs):
    """Helper function to translate between ambiguously named parameters

    This functions exists to help with backwards compatibility. Confusion exists around
    the true name of the "host", "port", and "url_prefix" parameters. Since this only
    applies when passing kwargs directly (as opposed to other config sources) we need to
    do the translation before asking yapconf to load the config.

    This function will:
    - Convert "host" to "bg_host"
    - Convert "port" to "bg_port"
    - Convert "url_prefix" to "bg_url_prefix"

    Args:
        **kwargs: Keyword arguments to translate
    """
    name_map = [
        # (old_name, new_name, show_deprecation)
        ("host", "bg_host", True),
        ("port", "bg_port", True),
        ("url_prefix", "bg_url_prefix", False),
    ]

    for old_name, new_name, show_deprecation in name_map:
        if old_name in kwargs:
            old_value = kwargs.pop(old_name)

            if new_name in kwargs:
                logger.warning(
                    "Looks like you passed both the '%s' and '%s' keyword arguments. "
                    "As these are aliases only the value from '%s' will be used."
                    % (old_name, new_name, new_name)
                )
            else:
                if show_deprecation:
                    _deprecate(
                        "Looks like you're using the '%s' keyword argument. This name "
                        "will be removed in version 4.0, please use '%s' instead."
                        % (old_name, new_name)
                    )

                kwargs[new_name] = old_value

    return kwargs



# -*- coding: utf-8 -*-
from brewtils.__version__ import __version__
from brewtils.config import get_argument_parser, get_connection_info, load_config
from brewtils.decorators import client, command, parameter, system
from brewtils.log import configure_logging
from brewtils.plugin import Plugin, RemotePlugin  # noqa F401
from brewtils.rest import normalize_url_prefix
from brewtils.rest.easy_client import get_easy_client, EasyClient
from brewtils.rest.system_client import SystemClient

__all__ = [
    "__version__",
    "client",
    "command",
    "parameter",
    "system",
    "Plugin",
    "EasyClient",
    "SystemClient",
    "get_easy_client",
    "get_argument_parser",
    "get_connection_info",
    "load_config",
    "configure_logging",
    "normalize_url_prefix",
]

# Aliased for compatibility
generated_version = __version__


# Alias old name for compatibility
def get_bg_connection_parameters(*args, **kwargs):
    from brewtils.errors import _deprecate

    _deprecate("get_bg_connection_parameters has been renamed to get_connection_info")
    return get_connection_info(*args, **kwargs)



# -*- coding: utf-8 -*-
from typing import Callable, Iterable, Optional, Union

import six
from lark import Lark, Transformer

# Lark added some new errors in later versions
# Lark also moved around their error in 0.6.0
from brewtils.errors import PluginParamError
from brewtils.models import Choices

try:
    from lark import ParseError
except ImportError:
    from lark.common import ParseError

try:
    from lark import GrammarError, LexError
except ImportError:
    GrammarError = ParseError
    LexError = ParseError


choices_grammar = r"""
    func: CNAME [func_args]
    url: ADDRESS [url_args]
    reference: ref

    func_args: "(" [arg_pair ("," arg_pair)*] ")"
    url_args: "?" arg_pair ("&" arg_pair)*

    arg_pair: CNAME "=" ref
    ?ref: "${" CNAME "}"

    ADDRESS: /^http[^\?]*/

    %import common.CNAME
    %import common.WS
    %ignore WS
"""

parsers = {
    "func": Lark(choices_grammar, start="func"),
    "url": Lark(choices_grammar, start="url"),
    "reference": Lark(choices_grammar, start="reference"),
}


class FunctionTransformer(Transformer):
    @staticmethod
    def func(s):
        return {"name": str(s[0]), "args": s[1] if len(s) > 1 else []}

    @staticmethod
    def url(s):
        return {"address": str(s[0]), "args": s[1] if len(s) > 1 else []}

    @staticmethod
    def reference(s):
        return str(s[0])

    @staticmethod
    def arg_pair(s):
        return str(s[0]), str(s[1])

    func_args = list
    url_args = list


def parse(input_string, parse_as=None):
    """Attempt to parse a string into a choices dictionary.

    Args:
        input_string: The string to parse
        parse_as: String specifying how to parse `input_string`. Valid values are
            'func' or 'url'. Will try all valid values if None.

    Returns:
        Dictionary containing the parse results

    Raises:
        lark.common.ParseError: Unable to find a valid parsing of `input_string`
    """

    def _parse(_input_string, _parser):
        try:
            return FunctionTransformer().transform(_parser.parse(_input_string))
        except (GrammarError, LexError, ParseError) as e:
            raise ParseError(e)

    if parse_as is not None:
        return _parse(input_string, parsers[parse_as])
    else:
        for parser in parsers.values():
            try:
                return _parse(input_string, parser)
            except ParseError:
                continue

    raise ParseError('Unable to successfully parse input "%s"' % input_string)


def _determine_display(display_value):
    if isinstance(display_value, six.string_types):
        return "typeahead"

    return "select" if len(display_value) <= 50 else "typeahead"


def _determine_type(type_value):
    if isinstance(type_value, six.string_types):
        return "url" if type_value.startswith("http") else "command"

    return "static"


def process_choices(choices):
    # type: (Union[dict, str, Iterable, Callable]) -> Optional[Choices]
    """Process a choices definition into a usable Choices object

    Args:
        choices: Raw choices definition, usually from a decorator

    Returns:
        Choices: Dictionary that fully describes a choices specification
    """

    if choices is None or isinstance(choices, Choices):
        return choices

    # If choices is a Callable, call it
    if callable(choices):
        choices = choices()

    if isinstance(choices, dict):
        if not choices.get("value"):
            raise PluginParamError(
                "No 'value' provided for choices. You must at least "
                "provide valid values."
            )

        # Again, if value is a Callable, call it
        value = choices.get("value")
        if callable(value):
            value = value()

        display = choices.get("display", _determine_display(value))
        choice_type = choices.get("type")
        strict = choices.get("strict", True)

        if choice_type is None:
            choice_type = _determine_type(value)
        elif choice_type not in Choices.TYPES:
            raise PluginParamError(
                "Invalid choices type '%s' - Valid type options are %s"
                % (choice_type, Choices.TYPES)
            )
        else:
            if (
                (
                    choice_type == "command"
                    and not isinstance(value, (six.string_types, dict))
                )
                or (choice_type == "url" and not isinstance(value, six.string_types))
                or (choice_type == "static" and not isinstance(value, (list, dict)))
            ):
                allowed_types = {
                    "command": "('string', 'dictionary')",
                    "url": "('string')",
                    "static": "('list', 'dictionary)",
                }
                raise PluginParamError(
                    "Invalid choices value type '%s' - Valid value types for "
                    "choice type '%s' are %s"
                    % (type(value), choice_type, allowed_types[choice_type])
                )

        if display not in Choices.DISPLAYS:
            raise PluginParamError(
                "Invalid choices display '%s' - Valid display options are %s"
                % (display, Choices.DISPLAYS)
            )

    elif isinstance(choices, str):
        value = choices
        display = _determine_display(value)
        choice_type = _determine_type(value)
        strict = True

    else:
        try:
            # Assume some sort of iterable
            value = list(choices)
        except TypeError:
            raise PluginParamError(
                "Invalid 'choices': must be a string, dictionary, or iterable."
            )

        display = _determine_display(value)
        choice_type = _determine_type(value)
        strict = True

    # Now parse out type-specific aspects
    unparsed_value = ""
    try:
        if choice_type == "command":
            if isinstance(value, six.string_types):
                unparsed_value = value
            else:
                unparsed_value = value["command"]

            details = parse(unparsed_value, parse_as="func")
        elif choice_type == "url":
            unparsed_value = value
            details = parse(unparsed_value, parse_as="url")
        else:
            if isinstance(value, dict):
                unparsed_value = choices.get("key_reference")
                if unparsed_value is None:
                    raise PluginParamError(
                        "Specifying a static choices dictionary requires a "
                        '"key_reference" field with a reference to another '
                        'parameter ("key_reference": "${param_key}")'
                    )

                details = {"key_reference": parse(unparsed_value, parse_as="reference")}
            else:
                details = {}
    except ParseError:
        raise PluginParamError(
            "Invalid choices definition - Unable to parse '%s'" % unparsed_value
        )

    return Choices(
        type=choice_type, display=display, value=value, strict=strict, details=details
    )



# -*- coding: utf-8 -*-
import abc
import json
import logging
import sys
import threading
from concurrent.futures.thread import ThreadPoolExecutor

import six
from requests import ConnectionError as RequestsConnectionError

import brewtils.plugin
from brewtils.errors import (
    BGGivesUpError,
    DiscardMessageException,
    RepublishRequestException,
    RequestProcessingError,
    RestClientError,
    RestConnectionError,
    TooLargeError,
    parse_exception_as_json,
)
from brewtils.models import Request
from brewtils.schema_parser import SchemaParser


class RequestProcessor(object):
    """Class responsible for coordinating Request processing

    The RequestProcessor is responsible for the following:
    - Defining on_message_received callback that will be invoked by the PikaConsumer
    - Parsing the request
    - Invoking the command on the target
    - Formatting the output
    - Reporting request updates to Beergarden (using a RequestUpdater)

    Args:
        target: Incoming requests will be invoked on this object
        updater: RequestUpdater that will be used for updating requests
        validation_funcs: List of functions that will called before invoking a command
        logger: A logger
        plugin_name: The Plugin's unique name
        max_workers: Max number of threads to use in the executor pool
    """

    def __init__(
        self,
        target,
        updater,
        consumer,
        validation_funcs=None,
        logger=None,
        plugin_name=None,
        max_workers=None,
        resolver=None,
        system=None,
    ):
        self.logger = logger or logging.getLogger(__name__)

        self.consumer = consumer
        self.consumer.on_message_callback = self.on_message_received

        self._target = target
        self._updater = updater
        self._plugin_name = plugin_name
        self._validation_funcs = validation_funcs or []
        self._pool = ThreadPoolExecutor(max_workers=max_workers)

        self._resolver = resolver
        self._system = system

    def on_message_received(self, message, headers):
        """Callback function that will be invoked for received messages

        This will attempt to parse the message and then run the parsed Request through
        all validation functions that this RequestProcessor knows about.

        If the request parses cleanly and passes validation it will be submitted to this
        RequestProcessor's ThreadPoolExecutor for processing.

        Args:
            message: The message string
            headers: The header dictionary

        Returns:
            A future that will complete when processing finishes

        Raises:
            DiscardMessageException: The request failed to parse correctly
            RequestProcessException: Validation failures should raise a subclass of this
        """
        request = self._parse(message)

        for func in self._validation_funcs:
            func(request)

        # This message has already been processed, all it needs to do is update
        if request.status in Request.COMPLETED_STATUSES:
            return self._pool.submit(self._updater.update_request, request, headers)
        else:
            return self._pool.submit(
                self.process_message, self._target, request, headers
            )

    def process_message(self, target, request, headers):
        """Process a message. Intended to be run on an Executor.

        Will set the status to IN_PROGRESS, invoke the command, and set the final
        status / output / error_class.

        Args:
            target: The object to invoke received commands on
            request: The parsed Request
            headers: Dictionary of headers from the `PikaConsumer`

        Returns:
            None
        """
        request.status = "IN_PROGRESS"
        self._updater.update_request(request, headers)

        try:
            # Set request context so this request will be the parent of any
            # generated requests and update status We also need the host/port of
            #  the current plugin. We currently don't support parent/child
            # requests across different servers.
            brewtils.plugin.request_context.current_request = request

            output = self._invoke_command(target, request, headers)
        except Exception as exc:
            self._handle_invoke_failure(request, exc)
        else:
            self._handle_invoke_success(request, output)

        self._updater.update_request(request, headers)

    def startup(self):
        """Start the RequestProcessor"""
        self.consumer.start()

    def shutdown(self):
        """Stop the RequestProcessor"""
        self.logger.debug("Shutting down consumer")
        self.consumer.stop_consuming()

        # Finish all current actions
        self._pool.shutdown(wait=True)

        self.consumer.stop()
        self.consumer.join()

        # Give the updater a chance to shutdown
        self._updater.shutdown()

    def _handle_invoke_success(self, request, output):
        request.status = "SUCCESS"
        request.output = self._format_output(output)

    def _handle_invoke_failure(self, request, exc):
        self.logger.log(
            getattr(exc, "_bg_error_log_level", logging.ERROR),
            "Plugin %s raised an exception while processing request %s: %s",
            self._plugin_name,
            str(request),
            exc,
            exc_info=not getattr(exc, "_bg_suppress_stacktrace", False),
        )
        request.status = "ERROR"
        request.output = self._format_error_output(request, exc)
        request.error_class = type(exc).__name__

    def _parse(self, message):
        """Parse a message using the standard SchemaParser

        Args:
            message: The raw (json) message body

        Returns:
            A Request model

        Raises:
            DiscardMessageException: The request failed to parse correctly
        """
        try:
            return SchemaParser.parse_request(message, from_string=True)
        except Exception as ex:
            self.logger.exception(
                "Unable to parse message body: {0}. Exception: {1}".format(message, ex)
            )
            raise DiscardMessageException("Error parsing message body")

    def _invoke_command(self, target, request, headers):
        """Invoke the function named in request.command

        Args:
            target: The object to search for the function implementation.
            request: The request to process
            headers: The headers for this request

        Returns:
            The output of the function call

        Raises:
            RequestProcessingError: The specified target does not define a
                callable implementation of request.command
        """
        if not callable(getattr(target, request.command, None)):
            raise RequestProcessingError(
                "Could not find an implementation of command '%s'" % request.command
            )

        # Get the command to use the parameter definitions when resolving
        command = None
        if self._system:
            command = self._system.get_command_by_name(request.command)

        # Now resolve parameters, if necessary
        if request.is_ephemeral or not command:
            parameters = request.parameters or {}
        else:
            parameters = self._resolver.resolve(
                request.parameters,
                definitions=command.parameters,
                upload=False,
            )

        return getattr(target, request.command)(**parameters)

    @staticmethod
    def _format_error_output(request, exc):
        if request.is_json:
            return parse_exception_as_json(exc)
        else:
            return str(exc)

    @staticmethod
    def _format_output(output):
        if isinstance(output, six.string_types):
            return output

        try:
            return json.dumps(output)
        except (TypeError, ValueError):
            return str(output)


class AdminProcessor(RequestProcessor):
    """RequestProcessor with slightly modified process method"""

    def process_message(self, target, request, headers):
        """Process a message. Intended to be run on an Executor.

        Will invoke the command and set the final status / output / error_class.

        Will NOT set the status to IN_PROGRESS or set the request context.

        Args:
            target: The object to invoke received commands on
            request: The parsed Request
            headers: Dictionary of headers from the `PikaConsumer`

        Returns:
            None
        """
        try:
            output = self._invoke_command(target, request, headers)
        except Exception as exc:
            self._handle_invoke_failure(request, exc)
        else:
            self._handle_invoke_success(request, output)

        self._updater.update_request(request, headers)


@six.add_metaclass(abc.ABCMeta)
class RequestConsumer(threading.Thread):
    """Base class for consumers

    Classes deriving from this are expected to provide a concrete implementation for a
    specific queue type.

    After the consumer is created it will be passed to a ``RequestProcessor``. The
    processor will then set the ``on_message_callback`` property of the consumer to the
    correct method.

    This means when the consumer receives a message it should invoke its own
    ``_on_message_callback`` method with the message body and headers as parameters::

        self._on_message_callback(body, properties.headers)

    """

    def __init__(self, *args, **kwargs):
        super(RequestConsumer, self).__init__(*args, **kwargs)
        self._on_message_callback = None

    def stop_consuming(self):
        pass

    def stop(self):
        pass

    @property
    def on_message_callback(self):
        return self._on_message_callback

    @on_message_callback.setter
    def on_message_callback(self, new_callback):
        self._on_message_callback = new_callback

    @staticmethod
    def create(connection_type=None, **kwargs):
        """Factory method for consumer creation

        Currently the only supported connection_type is "rabbitmq", which will return
        an instance of ``brewtils.pika.PikaConsumer``.

        Args:
            connection_type (str): String describing connection type
            kwargs: Keyword arguments to be passed to the Consumer initializer

        Returns:
            Concrete instance of RequestConsumer

        Raises:
            ValueError: The specified connection_type does not map to a consumer class
        """
        if connection_type == "rabbitmq":
            from brewtils.pika import PikaConsumer

            return PikaConsumer(**kwargs)

        raise ValueError("Unknown connection type '%s'" % connection_type)


@six.add_metaclass(abc.ABCMeta)
class RequestUpdater(object):
    @abc.abstractmethod
    def update_request(self, request, headers):
        pass

    @abc.abstractmethod
    def shutdown(self):
        pass


class NoopUpdater(RequestUpdater):
    """RequestUpdater implementation that explicitly does not update."""

    def __init__(self, *args, **kwargs):
        pass

    def update_request(self, request, headers):
        pass

    def shutdown(self):
        pass


class HTTPRequestUpdater(RequestUpdater):
    """RequestUpdater implementation based around an EasyClient.

    Args:
        ez_client: EasyClient to use for communication
        shutdown_event: `threading.Event` to allow for timely shutdowns
        logger: A logger

    Keyword Args:
        logger: A logger
        max_attempts: Max number of unsuccessful updates before discarding the message
        max_timeout: Maximum amount of time (seconds) to wait between update attempts
        starting_timeout: Starting time to wait (seconds) between update attempts

    """

    def __init__(self, ez_client, shutdown_event, **kwargs):
        self.logger = kwargs.get("logger", logging.getLogger(__name__))

        self._ez_client = ez_client
        self._shutdown_event = shutdown_event

        self.max_attempts = kwargs.get("max_attempts", -1)
        self.max_timeout = kwargs.get("max_timeout", 30)
        self.starting_timeout = kwargs.get("starting_timeout", 5)

        # Tightly manage when we're in an 'error' state, aka Brew-view is down
        self.brew_view_error_condition = threading.Condition()
        self.brew_view_down = False

        self.logger.debug("Creating and starting connection poll thread")
        self.connection_poll_thread = self._create_connection_poll_thread()
        self.connection_poll_thread.start()

    def shutdown(self):
        self.logger.debug("Shutting down, about to wake any sleeping updater threads")
        with self.brew_view_error_condition:
            self.brew_view_error_condition.notify_all()

    def update_request(self, request, headers):
        """Sends a Request update to beer-garden

        Ephemeral requests do not get updated, so we simply skip them.

        If brew-view appears to be down, it will wait for brew-view to come back
         up before updating.

        If this is the final attempt to update, we will attempt a known, good
        request to give some information to the user. If this attempt fails
        then we simply discard the message.

        Args:
            request: The request to update
            headers: A dictionary of headers from the `PikaConsumer`

        Returns:
            None

        Raises:
            RepublishMessageException: The Request update failed (any reason)
        """
        if request.is_ephemeral:
            sys.stdout.flush()
            return

        with self.brew_view_error_condition:

            self._wait_for_brew_view_if_down(request)

            try:
                if not self._should_be_final_attempt(headers):
                    self._wait_if_not_first_attempt(headers)
                    self._ez_client.update_request(
                        request.id,
                        status=request.status,
                        output=request.output,
                        error_class=request.error_class,
                    )
                else:
                    self._ez_client.update_request(
                        request.id,
                        status="ERROR",
                        output="We tried to update the request, but it failed too many "
                        "times. Please check the plugin logs to figure out why the "
                        "request update failed. It is possible for this request to "
                        "have succeeded, but we cannot update beer-garden with that "
                        "information.",
                        error_class="BGGivesUpError",
                    )
            except Exception as ex:
                self._handle_request_update_failure(request, headers, ex)
            finally:
                sys.stdout.flush()

    def _wait_if_not_first_attempt(self, headers):
        if headers.get("retry_attempt", 0) > 0:
            time_to_sleep = min(
                headers.get("time_to_wait", self.starting_timeout), self.max_timeout
            )
            self._shutdown_event.wait(time_to_sleep)

    def _handle_request_update_failure(self, request, headers, exc):

        # If brew-view is down, we always want to try again
        # Yes, even if it is the 'final_attempt'
        if isinstance(exc, (RequestsConnectionError, RestConnectionError)):
            self.brew_view_down = True
            self.logger.error(
                "Error updating request status: {0} exception: {1}".format(
                    request.id, exc
                )
            )
            raise RepublishRequestException(request, headers)

        elif isinstance(exc, TooLargeError):
            self.logger.error(
                "Error updating request {0} - the request exceeds the 16MB size "
                "limitation. The status of this request will be marked as ERROR, but "
                "it's possible the request actually completed successfully. If this "
                "happens often please contact the plugin developer.".format(request.id)
            )

            request.status = "ERROR"
            request.output = "Request size greater than 16MB"
            request.error_class = BGGivesUpError.__name__

            raise RepublishRequestException(request, headers)

        elif isinstance(exc, RestClientError):
            message = (
                "Error updating request {0} and it is a client error. Probable "
                "cause is that this request is already updated. In which case, "
                "ignore this message. If request {0} did not complete, please "
                "file an issue. Discarding request to avoid an infinite loop. "
                "exception: {1}".format(request.id, exc)
            )
            self.logger.error(message)
            raise DiscardMessageException(message)

        # Time to discard the message because we've given up
        elif self._should_be_final_attempt(headers):
            message = (
                "Could not update request {0} even with a known good status, "
                "output and error_class. We have reached the final attempt and "
                "will now discard the message. Attempted to process this "
                "message {1} times".format(request.id, headers["retry_attempt"])
            )
            self.logger.error(message)
            raise DiscardMessageException(message)

        else:
            self._update_retry_attempt_information(headers)
            self.logger.exception(
                "Error updating request (Attempt #{0}: request: {1} exception: "
                "{2}".format(headers.get("retry_attempt", 0), request.id, exc)
            )
            raise RepublishRequestException(request, headers)

    def _update_retry_attempt_information(self, headers):
        headers["retry_attempt"] = headers.get("retry_attempt", 0) + 1
        headers["time_to_wait"] = min(
            headers.get("time_to_wait", self.starting_timeout // 2) * 2,
            self.max_timeout,
        )

    def _should_be_final_attempt(self, headers):
        if self.max_attempts <= 0:
            return False

        return self.max_attempts <= headers.get("retry_attempt", 0)

    def _wait_for_brew_view_if_down(self, request):
        if self.brew_view_down and not self._shutdown_event.is_set():
            self.logger.warning(
                "Currently unable to communicate with Brew-view, about to wait "
                "until connection is reestablished to update request %s",
                request.id,
            )
            self.brew_view_error_condition.wait()

    def _create_connection_poll_thread(self):
        connection_poll_thread = threading.Thread(target=self._connection_poll)
        connection_poll_thread.daemon = True
        return connection_poll_thread

    def _connection_poll(self):
        """Periodically attempt to re-connect to beer-garden"""

        while not self._shutdown_event.wait(5):
            try:
                with self.brew_view_error_condition:
                    if self.brew_view_down:
                        try:
                            self._ez_client.get_version()
                        except Exception:
                            self.logger.debug("Brew-view reconnection attempt failure")
                        else:
                            self.logger.info(
                                "Brew-view connection reestablished, about to "
                                "notify any waiting requests"
                            )
                            self.brew_view_down = False
                            self.brew_view_error_condition.notify_all()
            except Exception as ex:
                self.logger.exception("Exception in connection poll thread: %s", ex)



# -*- coding: utf-8 -*-
"""Brewtils Logging Utilities

This module streamlines loading logging configuration from Beergarden.

Example:
    To use this just call ``configure_logging`` sometime before you initialize
    your Plugin object:

    .. code-block:: python

        from brewtils import configure_logging, get_connection_info, Plugin

        # Load BG connection info from environment and command line args
        connection_info = get_connection_info(sys.argv[1:])

        configure_logging(system_name='systemX', **connection_info)

        plugin = Plugin(
            my_client,
            name='systemX,
            version='0.0.1',
            **connection_info
        )
        plugin.run()
"""

import copy
import json
import os
import re
import string
import warnings

import logging.config

import brewtils

DEFAULT_LOGGERS = {
    "pika": {"level": "ERROR"},
    "requests.packages.urllib3.connectionpool": {"level": "WARN"},
    "yapconf": {"level": "WARN"},
}

DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_FORMATTERS = {"default": {"format": DEFAULT_FORMAT}}

DEFAULT_HANDLERS = {
    "default": {
        "class": "logging.StreamHandler",
        "formatter": "default",
        "stream": "ext://sys.stdout",
    }
}

DEFAULT_ROOT = {"level": "INFO", "formatter": "default", "handlers": ["default"]}

DEFAULT_PLUGIN_LOGGING_TEMPLATE = {
    "version": 1,
    "disable_existing_loggers": False,
    "loggers": DEFAULT_LOGGERS,
    "formatters": DEFAULT_FORMATTERS,
    "handlers": DEFAULT_HANDLERS,
    "root": DEFAULT_ROOT,
}


def default_config(level="INFO"):
    """Get a basic logging configuration with the given level"""
    config = copy.deepcopy(DEFAULT_PLUGIN_LOGGING_TEMPLATE)
    config["root"]["level"] = level

    return config


def configure_logging(
    raw_config,
    namespace=None,
    system_name=None,
    system_version=None,
    instance_name=None,
):
    """Load and enable a logging configuration from Beergarden

    WARNING: This method will modify the current logging configuration.

    The configuration will be template substituted using the keyword arguments passed
    to this function. For example, a handler like this:

    .. code-block:: yaml

        handlers:
            file:
                backupCount: 5
                class: "logging.handlers.RotatingFileHandler"
                encoding: utf8
                formatter: default
                level: INFO
                maxBytes: 10485760
                filename: "$system_name.log"

    Will result in logging to a file with the same name as the given system_name.

    This will also ensure that directories exist for any file-based handlers. Default
    behavior for the Python logging module is to not create directories that do not
    already exist, which would dramatically lower the utility of templating.

    Args:
        raw_config: Configuration to apply
        namespace: Used for configuration templating
        system_name: Used for configuration templating
        system_version: Used for configuration templating
        instance_name: Used for configuration templating

    Returns:
        None
    """

    class ConfigParserTemplate(string.Template):
        """string.Template variant for ConfigParser-style interpolation

        So. This exists because we want to do template substitution on the logging
        configuration file. We want this to be consistent with how the logging module
        itself does substitution, and since we need this to work on Python 2 that means
        the ConfigParser flavor: %(variable)s

        The important parts here that differ from the normal string.Template are:
        - The delimiter ("%" instead of "$")
        - The "delimiter and a braced identifier" part of the pattern definition. This
          is needed to match %(variable)s instead of %{variable} like a normal template
        - The "id" and additional field "bid" in Python 3.7 are slightly different:
          r"(?a:[_a-z][_a-z0-9]*)" instead of r"[_a-z][_a-z0-9]*"
          Hopefully that's not a problem.
        """

        delimiter = "%"

        pattern = r"""
        %(delim)s(?:
          (?P<escaped>%(delim)s)    |   # Escape sequence of two delimiters
          (?P<named>%(id)s)         |   # delimiter and a Python identifier
          \((?P<braced>%(id)s)\)s  |   # delimiter and a braced identifier
          (?P<invalid>)                 # Other ill-formed delimiter exprs
        )
        """ % {
            "delim": re.escape("%"),
            "id": r"[_a-z][_a-z0-9]*",
        }

    templated = ConfigParserTemplate(json.dumps(raw_config)).safe_substitute(
        namespace=namespace,
        system_name=system_name,
        system_version=system_version,
        instance_name=instance_name,
    )
    logging_config = json.loads(templated)

    # Now make sure that directories for all file handlers exist
    for handler in logging_config["handlers"].values():
        if "filename" in handler:
            dir_name = os.path.dirname(os.path.abspath(handler["filename"]))
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)

    logging.config.dictConfig(logging_config)


def find_log_file():
    """Find the file name for the first file handler attached to the root logger"""
    for h in logging.getLogger().handlers:
        if hasattr(h, "baseFilename"):
            return h.baseFilename


def read_log_file(log_file, start_line=None, end_line=None):
    """Read lines from a log file

    Args:
        log_file: The file to read from
        start_line: Starting line to read
        end_line: Ending line to read

    Returns:
        Lines read from the file
    """
    with open(log_file, "r") as f:
        raw_logs = f.readlines()

    return "".join(raw_logs[start_line:end_line])


# DEPRECATED
SUPPORTED_HANDLERS = ("stdout", "file", "logstash")


def get_logging_config(system_name=None, **kwargs):
    """Retrieve a logging configuration from Beergarden

    Args:
        system_name: Name of the system to load
        **kwargs: Beergarden connection parameters

    Returns:
        dict: The logging configuration for the specified system
    """
    warnings.warn(
        "This function is deprecated and will be removed in version "
        "4.0, please consider using 'EasyClient.get_logging_config' and "
        "'configure_logging' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    config = brewtils.get_easy_client(**kwargs).get_logging_config(system_name)

    return convert_logging_config(config)


def convert_logging_config(logging_config):
    """Transform a LoggingConfig object into a Python logging configuration

    Args:
        logging_config: Beergarden logging config

    Returns:
        dict: The logging configuration
    """
    warnings.warn(
        "This function is deprecated and will be removed in version "
        "4.0, please consider using 'configure_logging' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    config_to_return = copy.deepcopy(DEFAULT_PLUGIN_LOGGING_TEMPLATE)

    if logging_config.handlers:
        handlers = logging_config.handlers
    else:
        handlers = copy.deepcopy(DEFAULT_HANDLERS)
    config_to_return["handlers"] = handlers

    if logging_config.formatters:
        formatters = logging_config.formatters
    else:
        formatters = copy.deepcopy(DEFAULT_FORMATTERS)
    config_to_return["formatters"] = formatters

    config_to_return["root"] = {
        "level": logging_config.level,
        "handlers": list(config_to_return["handlers"]),
    }

    return config_to_return


def setup_logger(
    bg_host, bg_port, system_name, ca_cert=None, client_cert=None, ssl_enabled=None
):
    """DEPRECATED: Set Python logging to use configuration from Beergarden API

    This method is deprecated - consider using :func:`configure_logging`

    This method will overwrite the current logging configuration.

    Args:
        bg_host (str): Beergarden host
        bg_port (int): Beergarden port
        system_name (str): Name of the system
        ca_cert (str): Path to CA certificate file
        client_cert (str): Path to client certificate file
        ssl_enabled (bool): Use SSL when connection to Beergarden

    Returns: None
    """
    warnings.warn(
        "This function is deprecated and will be removed in version "
        "4.0, please consider using 'configure_logging' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    config = get_python_logging_config(
        bg_host=bg_host,
        bg_port=bg_port,
        system_name=system_name,
        ca_cert=ca_cert,
        client_cert=client_cert,
        ssl_enabled=ssl_enabled,
    )
    logging.config.dictConfig(config)


def get_python_logging_config(
    bg_host, bg_port, system_name, ca_cert=None, client_cert=None, ssl_enabled=None
):
    """DEPRECATED: Get Beergarden's logging configuration

    This method is deprecated - consider using :func:`get_logging_config`

    Args:
        bg_host (str): Beergarden host
        bg_port (int): Beergarden port
        system_name (str): Name of the system
        ca_cert (str): Path to CA certificate file
        client_cert (str): Path to client certificate file
        ssl_enabled (bool): Use SSL when connection to Beergarden

    Returns:
        dict: The logging configuration for the specified system
    """
    warnings.warn(
        "This function is deprecated and will be removed in version "
        "4.0, please consider using 'get_logging_config' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    client = brewtils.get_easy_client(
        host=bg_host,
        port=bg_port,
        ssl_enabled=ssl_enabled,
        ca_cert=ca_cert,
        client_cert=client_cert,
    )

    logging_config = client.get_logging_config(system_name=system_name)

    return convert_logging_config(logging_config)



# -*- coding: utf-8 -*-
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from multiprocessing import cpu_count
from typing import Any, Dict, Iterable, Optional

from packaging.version import parse

import brewtils.plugin
from brewtils.errors import (
    FetchError,
    RequestFailedError,
    RequestProcessException,
    TimeoutExceededError,
    ValidationError,
    _deprecate,
)
from brewtils.models import Request, System
from brewtils.resolvers.manager import ResolutionManager
from brewtils.rest.easy_client import EasyClient


class SystemClient(object):
    """High-level client for generating requests for a Beer-garden System.

    SystemClient creation:
        This class is intended to be the main way to create Beer-garden requests. Create
        an instance with Beer-garden connection information and a system name::

            client = SystemClient(
                system_name='example_system',
                system_namespace='default',
                bg_host="host",
                bg_port=2337,
            )

        Note: Passing an empty string as the system_namespace parameter will evalutate
        to the local garden's default namespace.

        Pass additional keyword arguments for more granularity:

            version_constraint:
                Allows specifying a particular system version. Can be a version literal
                ('1.0.0') or the special value 'latest.' Using 'latest' will allow the
                SystemClient to retry a request if it fails due to a missing system
                (see Creating Requests).

            default_instance:
                The instance name to use when creating a request if no other instance
                name is specified. Since each request must be addressed to a specific
                instance this is a convenience to prevent needing to specify the
                instance for each request.

            always_update:
                If True the SystemClient will always attempt to reload the system
                definition before making a request. This is useful to ensure Requests
                are always made against the latest version of the system.
                If not set the System definition will be loaded when making the first
                request and will only be reloaded if a Request fails.

    Loading the System:
        The System definition is lazily loaded, so nothing happens until the first
        attempt to send a Request. At that point the SystemClient will query Beer-garden
        to get a system definition that matches the system_name and version_constraint.
        If no matching System can be found a FetchError will be raised. If always_update
        was set to True this will happen before making each request, not only the first.

    Making a Request:
        The standard way to create and send requests is by calling object attributes::

            request = client.example_command(param_1='example_param')

        In the normal case this will block until the request completes. Request
        completion is determined by periodically polling Beer-garden to check the
        Request status. The time between polling requests starts at 0.5s and doubles
        each time the request has still not completed, up to max_delay. If a timeout was
        specified and the Request has not completed within that time a
        ``ConnectionTimeoutError`` will be raised.

        It is also possible to create the SystemClient in non-blocking mode by
        specifying blocking=False. In this case the request creation will immediately
        return a Future and will spawn a separate thread to poll for Request completion.
        The max_concurrent parameter is used to control the maximum threads available
        for polling.

        .. code-block:: python

            # Create a SystemClient with blocking=False
            client = SystemClient(
                system_name='example_system',
                system_namespace='default',
                bg_host="localhost",
                bg_port=2337,
                blocking=False,
            )

            # Create and send 5 requests without waiting for request completion
            futures = [client.example_command(param_1=number) for number in range(5)]

            # Now wait on all requests to complete
            concurrent.futures.wait(futures)

        If the request creation process fails (e.g. the command failed validation) and
        version_constraint is 'latest' then the SystemClient will check to see if a
        newer version is available, and if so it will attempt to make the request on
        that version. This is so users of the SystemClient that don't necessarily care
        about the target system version don't need to be restarted every time the target
        system is updated.

        It's also possible to control what happens when a Request results in an ERROR.
        If the ``raise_on_error`` parameter is set to False (the default) then Requests
        that are not successful simply result in a Request with a status of ERROR, and
        it is the plugin developer's responsibility to check for this case. However, if
        ``raise_on_error`` is set to True then this will result in a
        ``RequestFailedError`` being raised. This will happen regardless of the value
        of the ``blocking`` flag.

    Tweaking Beer-garden Request Parameters:
        There are several parameters that control how beer-garden routes / processes a
        request. To denote these as intended for Beer-garden itself (rather than a
        parameter to be passed to the Plugin) prepend a leading underscore to the
        argument name.

        Sending to another instance::

            request = client.example_command(
                _instance_name="instance_2", param_1="example_param"
            )

        Request with a comment::

            request = client.example_command(
                _comment="I'm a beer-garden comment!", param_1="example_param"
            )

        Without the leading underscore the arguments would be treated the same as
        "param_1" - another parameter to be passed to the plugin.

        Request that raises::

            client = SystemClient(
                system_name="foo",
                system_namespace='default',
                bg_host="localhost",
                bg_port=2337,
            )

            try:
                client.command_that_errors(_raise_on_error=True)
            except RequestFailedError:
                print("I could have just ignored this")

    Args:
        system_name (str): Name of the System to make Requests on
        system_namespace (str): Namespace of the System to make Requests on
        version_constraint (str): System version to make Requests on. Can be specific
            ('1.0.0') or 'latest'.
        default_instance (str): Name of the Instance to make Requests on
        always_update (bool): Whether to check if a newer version of the System exists
            before making each Request. Only relevant if ``version_constraint='latest'``
        timeout (int): Seconds to wait for a request to complete. 'None' means wait
            forever.
        max_delay (int): Maximum number of seconds to wait between status checks for a
            created request
        blocking (bool): Flag indicating whether creation will block until the Request
            is complete or return a Future that will complete when the Request does
        max_concurrent (int): Maximum number of concurrent requests allowed.
            Only has an effect when blocking=False.
        raise_on_error (bool): Flag controlling whether created Requests that complete
            with an ERROR state should raise an exception

        bg_host (str): Beer-garden hostname
        bg_port (int): Beer-garden port
        bg_url_prefix (str): URL path that will be used as a prefix when communicating
            with Beer-garden. Useful if Beer-garden is running on a URL other than '/'.
        ssl_enabled (bool): Whether to use SSL for Beer-garden communication
        ca_cert (str): Path to certificate file containing the certificate of the
            authority that issued the Beer-garden server certificate
        ca_verify (bool): Whether to verify Beer-garden server certificate
        client_cert (str): Path to client certificate to use when communicating with
            Beer-garden
        api_version (int): Beer-garden API version to use
        client_timeout (int): Max time to wait for Beer-garden server response
        username (str): Username for Beer-garden authentication
        password (str): Password for Beer-garden authentication
        access_token (str): Access token for Beer-garden authentication
        refresh_token (str): Refresh token for Beer-garden authentication
    """

    def __init__(self, *args, **kwargs):
        self._logger = logging.getLogger(__name__)

        self._loaded = False
        self._system = None
        self._commands = {}

        # Need this for back-compatibility (see #836)
        if len(args) > 2:
            _deprecate(
                "Heads up - passing system_name as a positional argument is deprecated "
                "and will be removed in version 4.0",
            )
            kwargs.setdefault("system_name", args[2])

        # Now need to determine if the intended target is the current running plugin.
        # Start by ensuring there's a valid Plugin context active
        target_self = bool(brewtils.plugin.CONFIG)

        # If ANY of the target specification arguments don't match the current plugin
        # then the target is different
        config_map = {
            "system_name": "name",
            "version_constraint": "version",
            "default_instance": "instance_name",
            "system_namespace": "namespace",
        }
        for key, value in config_map.items():
            if (
                kwargs.get(key) is not None
                and kwargs.get(key) != brewtils.plugin.CONFIG[value]
            ):
                target_self = False
                break

        # Now assign self._system_name, etc based on the value of target_self
        if target_self:
            self._system_name = brewtils.plugin.CONFIG.name
            self._version_constraint = brewtils.plugin.CONFIG.version
            self._default_instance = brewtils.plugin.CONFIG.instance_name
            self._system_namespace = brewtils.plugin.CONFIG.namespace or ""
        else:
            self._system_name = kwargs.get("system_name")
            self._version_constraint = kwargs.get("version_constraint", "latest")
            self._default_instance = kwargs.get("default_instance", "default")
            self._system_namespace = kwargs.get(
                "system_namespace", brewtils.plugin.CONFIG.namespace or ""
            )

        self._always_update = kwargs.get("always_update", False)
        self._timeout = kwargs.get("timeout", None)
        self._max_delay = kwargs.get("max_delay", 30)
        self._blocking = kwargs.get("blocking", True)
        self._raise_on_error = kwargs.get("raise_on_error", False)

        # This is for Python 3.4 compatibility - max_workers MUST be non-None
        # in that version. This logic is what was added in Python 3.5
        max_concurrent = kwargs.get("max_concurrent", (cpu_count() or 1) * 5)
        self._thread_pool = ThreadPoolExecutor(max_workers=max_concurrent)

        # This points DeprecationWarnings at the right line
        kwargs.setdefault("stacklevel", 5)

        self._easy_client = EasyClient(*args, **kwargs)
        self._resolver = ResolutionManager(easy_client=self._easy_client)

    def __getattr__(self, item):
        # type: (str) -> partial
        """Standard way to create and send beer-garden requests"""
        return self.create_bg_request(item)

    def __str__(self):
        return "%s[%s]" % (self.bg_system, self.bg_default_instance)

    @property
    def bg_system(self):
        return self._system

    @property
    def bg_default_instance(self):
        return self._default_instance

    def create_bg_request(self, command_name, **kwargs):
        # type: (str, **Any) -> partial
        """Create a callable that will execute a Beer-garden request when called.

        Normally you interact with the SystemClient by accessing attributes, but there
        could be certain cases where you want to create a request without sending it.

        Example::

            client = SystemClient(host, port, 'system', blocking=False)

            # Create two callables - one with a parameter and one without
            uncreated_requests = [
                client.create_bg_request('command_1', arg_1='Hi!'),
                client.create_bg_request('command_2'),
            ]

            # Calling creates and sends the request
            # The result of each is a future because blocking=False on the SystemClient
            futures = [req() for req in uncreated_requests]

            # Wait for all the futures to complete
            concurrent.futures.wait(futures)

        Args:
            command_name (str): Name of the Command to send
            kwargs (dict): Will be passed as parameters when creating the Request

        Returns:
            Partial that will create and execute a Beer-garden request when called

        Raises:
            AttributeError: System does not have a Command with the given command_name
        """

        if not self._loaded or self._always_update:
            self.load_bg_system()

        if command_name in self._commands:
            return partial(
                self.send_bg_request,
                _command=command_name,
                _system_name=self._system.name,
                _system_namespace=self._system.namespace,
                _system_version=self._system.version,
                _system_display=self._system.display_name,
                _output_type=self._commands[command_name].output_type,
                _instance_name=self._default_instance,
                **kwargs
            )
        else:
            raise AttributeError(
                "System '%s' has no command named '%s'" % (self._system, command_name)
            )

    def send_bg_request(self, *args, **kwargs):
        """Actually create a Request and send it to Beer-garden

        .. note::
            This method is intended for advanced use only, mainly cases where you're
            using the SystemClient without a predefined System. It assumes that
            everything needed to construct the request is being passed in ``kwargs``. If
            this doesn't sound like what you want you should check out
            ``create_bg_request``.

        Args:
            args (list): Unused. Passing positional parameters indicates a bug
            kwargs (dict): All necessary request parameters, including Beer-garden
                internal parameters

        Returns:
            blocking=True: A completed Request object
            blocking=False: A future that will be completed when the Request does

        Raises:
            ValidationError: Request creation failed validation on the server
        """
        # First, if any positional args were given that's a bug, as it means someone
        # tried to pass a parameter without a key:
        # client.command_name(param)
        if args:
            raise RequestProcessException(
                "Using positional arguments when creating a request is not allowed. "
                "Please use keyword arguments instead."
            )

        # Need to pop here, otherwise we'll try to send as a request parameter
        raise_on_error = kwargs.pop("_raise_on_error", self._raise_on_error)
        blocking = kwargs.pop("_blocking", self._blocking)
        timeout = kwargs.pop("_timeout", self._timeout)

        # If the request fails validation and the version constraint allows,
        # check for a new version and retry
        try:
            request = self._construct_bg_request(**kwargs)
            request = self._easy_client.create_request(
                request, blocking=blocking, timeout=timeout
            )
        except ValidationError:
            if self._system and self._version_constraint == "latest":
                old_version = self._system.version

                self.load_bg_system()

                if old_version != self._system.version:
                    kwargs["_system_version"] = self._system.version
                    return self.send_bg_request(**kwargs)
            raise

        # If not blocking just return the future
        if not blocking:
            return self._thread_pool.submit(
                self._wait_for_request, request, raise_on_error, timeout
            )

        # Brew-view before 2.4 doesn't support the blocking flag, so make sure
        # the request is actually complete before returning
        return self._wait_for_request(request, raise_on_error, timeout)

    def load_bg_system(self):
        # type: () -> None
        """Query beer-garden for a System definition

        This method will make the query to beer-garden for a System matching the name
        and version constraints specified during SystemClient instance creation.

        If this method completes successfully the SystemClient will be ready to create
        and send Requests.

        Returns:
            None

        Raises:
            FetchError: Unable to find a matching System
        """

        if self._version_constraint == "latest":
            self._system = self._determine_latest(
                self._easy_client.find_systems(
                    name=self._system_name, namespace=self._system_namespace
                )
            )
        else:
            self._system = self._easy_client.find_unique_system(
                name=self._system_name,
                version=self._version_constraint,
                namespace=self._system_namespace,
            )

        if self._system is None:
            raise FetchError(
                "Beer-garden has no system named '%s' with a version matching '%s' in "
                "namespace '%s'"
                % (
                    self._system_name,
                    self._version_constraint,
                    self._system_namespace
                    if self._system_namespace
                    else "<garden default>",
                )
            )

        self._commands = {command.name: command for command in self._system.commands}
        self._loaded = True

    def _wait_for_request(self, request, raise_on_error, timeout):
        # type: (Request, bool, int) -> Request
        """Poll the server until the request is completed or errors"""

        delay_time = 0.5
        total_wait_time = 0
        while request.status not in Request.COMPLETED_STATUSES:

            if timeout and 0 < timeout < total_wait_time:
                raise TimeoutExceededError(
                    "Timeout waiting for request '%s' to complete" % str(request)
                )

            time.sleep(delay_time)
            total_wait_time += delay_time
            delay_time = min(delay_time * 2, self._max_delay)

            request = self._easy_client.find_unique_request(id=request.id)

        if raise_on_error and request.status == "ERROR":
            raise RequestFailedError(request)

        return request

    def _get_parent_for_request(self):
        # type: () -> Optional[Request]
        parent = getattr(brewtils.plugin.request_context, "current_request", None)
        if parent is None:
            return None

        if brewtils.plugin.CONFIG and (
            brewtils.plugin.CONFIG.bg_host.upper()
            != self._easy_client.client.bg_host.upper()
            or brewtils.plugin.CONFIG.bg_port != self._easy_client.client.bg_port
        ):
            self._logger.warning(
                "A parent request was found, but the destination beer-garden "
                "appears to be different than the beer-garden to which this plugin "
                "is assigned. Cross-server parent/child requests are not supported "
                "at this time. Removing the parent context so the request doesn't fail."
            )
            return None

        return Request(id=str(parent.id))

    def _construct_bg_request(self, **kwargs):
        # type: (**Any) -> Request
        """Create a request that can be used with EasyClient.create_request"""

        command = kwargs.pop("_command", None)
        system_name = kwargs.pop("_system_name", None)
        system_version = kwargs.pop("_system_version", None)
        system_display = kwargs.pop("_system_display", None)
        system_namespace = kwargs.pop("_system_namespace", None)
        instance_name = kwargs.pop("_instance_name", None)
        comment = kwargs.pop("_comment", None)
        output_type = kwargs.pop("_output_type", None)
        metadata = kwargs.pop("_metadata", {})
        parent = kwargs.pop("_parent", self._get_parent_for_request())

        if system_display:
            metadata["system_display_name"] = system_display

        # Don't check namespace - https://github.com/beer-garden/beer-garden/issues/827
        if command is None:
            raise ValidationError("Unable to send a request with no command")
        if system_name is None:
            raise ValidationError("Unable to send a request with no system name")
        if system_version is None:
            raise ValidationError("Unable to send a request with no system version")
        if instance_name is None:
            raise ValidationError("Unable to send a request with no instance name")

        request = Request(
            command=command,
            system=system_name,
            system_version=system_version,
            namespace=system_namespace,
            instance_name=instance_name,
            comment=comment,
            output_type=output_type,
            parent=parent,
            metadata=metadata,
            parameters=kwargs,
        )

        request.parameters = self._resolve_parameters(command, request)

        return request

    def _resolve_parameters(self, command, request):
        # type: (str, Request) -> Dict[str, Any]
        """Attempt to upload any necessary file parameters

        This will inspect the Command model for the given command, looking for file
        parameters. Any file parameters will be "resolved" (aka uploaded) before
        continuing.

        If the command name can not be found in the current list of commands the
        parameter list will just be returned. This most likely indicates a direct
        invocation of send_bg_request since a bad command name should be caught earlier
        in the "normal" workflow.
        """
        if command not in self._commands:
            return request.parameters

        return self._resolver.resolve(
            request.parameters, self._commands[command].parameters, upload=True
        )

    @staticmethod
    def _determine_latest(systems):
        # type: (Iterable[System]) -> Optional[System]
        return (
            sorted(systems, key=lambda x: parse(x.version), reverse=True)[0]
            if systems
            else None
        )



# -*- coding: utf-8 -*-


def _is_json_dict(s):
    import json

    try:
        return isinstance(json.loads(s), dict)
    except json.decoder.JSONDecodeError:
        return False


_CONNECTION_SPEC = {
    "bg_host": {
        "type": "str",
        "description": "Beergarden server FQDN",
        "required": True,
        "env_name": "HOST",
        "alt_env_names": ["WEB_HOST"],
    },
    "bg_port": {
        "type": "int",
        "description": "Beergarden server port",
        "default": 2337,
        "env_name": "PORT",
        "alt_env_names": ["WEB_PORT"],
    },
    "bg_url_prefix": {
        "type": "str",
        "description": "Beergarden server path",
        "default": "/",
        "env_name": "URL_PREFIX",
        "cli_name": "url_prefix",
    },
    "ca_cert": {
        "type": "str",
        "description": "CA certificate to use when verifying",
        "required": False,
        "alt_env_names": ["SSL_CA_CERT"],
    },
    "ca_verify": {
        "type": "bool",
        "description": "Verify server certificate when using SSL",
        "default": True,
    },
    "client_cert": {
        "type": "str",
        "description": "Client certificate to use with Beergarden",
        "required": False,
        "alt_env_names": ["SSL_CLIENT_CERT"],
    },
    "client_key": {
        "type": "str",
        "description": "Client key to use with Beergarden",
        "required": False,
    },
    "ssl_enabled": {
        "type": "bool",
        "description": "Use SSL when communicating with Beergarden",
        "default": True,
    },
    "api_version": {
        "type": "int",
        "description": "Beergarden API version",
        "default": 1,
        "choices": [1],
    },
    "client_timeout": {
        "type": "float",
        "description": "Max time RestClient will wait for server response",
        "long_description": "This setting controls how long the HTTP(s) client will "
        "wait when opening a connection to Beergarden before aborting."
        "This prevents some strange Beergarden server state from causing "
        "plugins to hang indefinitely."
        "Set to -1 to disable (this is a bad idea in production code, see "
        "the Requests documentation).",
        "default": -1,
    },
    "username": {
        "type": "str",
        "description": "Username for authentication",
        "required": False,
    },
    "password": {
        "type": "str",
        "description": "Password for authentication",
        "required": False,
    },
    "access_token": {
        "type": "str",
        "description": "Access token for authentication",
        "required": False,
    },
    "refresh_token": {
        "type": "str",
        "description": "Refresh token for authentication",
        "required": False,
    },
    "proxy": {
        "type": "str",
        "description": "HTTP proxy to use when communicating with Beergarden",
        "long_description": "Proxy information should be given in the format "
        "of `<hostname>:<port>`. This setting will overwrite default proxy "
        "set by the HTTP(s)_PROXY environment variable. Use of SSL for this "
        "proxy will be determined by the ssl_enabled config option.",
        "default": "",
        "required": False,
    },
}

_SYSTEM_SPEC = {
    "name": {"type": "str", "description": "The system name", "required": False},
    "version": {"type": "str", "description": "The system version", "required": False},
    "description": {
        "type": "str",
        "description": "The system description",
        "required": False,
    },
    "max_instances": {
        "type": "int",
        "description": "The system max instances",
        "default": -1,
    },
    "icon_name": {
        "type": "str",
        "description": "The system icon name",
        "required": False,
    },
    "display_name": {
        "type": "str",
        "description": "The system display name",
        "required": False,
    },
    "metadata": {
        "type": "str",
        "description": "The system metadata, in JSON string form."
        'Something like \'{"foo": "bar"}\'',
        "default": "{}",
        "validator": _is_json_dict,
    },
    "namespace": {
        "type": "str",
        "description": "The namespace this system will be created in",
        "required": False,
    },
    "template": {
        "type": "str",
        "description": "Custom system page definition",
        "required": False,
        "long_description": "Follows the same rules as @command templates - can be "
        "a URL, path string to a template file, or an actual html string",
    },
}

_PLUGIN_SPEC = {
    "instance_name": {
        "type": "str",
        "description": "The instance name",
        "default": "default",
    },
    "runner_id": {
        "type": "str",
        "description": "The PluginRunner ID, if applicable",
        "required": False,
    },
    "log_level": {
        "type": "str",
        "description": "The log level to use",
        "default": "INFO",
        "bootstrap": True,
    },
    "max_concurrent": {
        "type": "int",
        "description": "Maximum number of requests to process concurrently",
        "default": 5,
    },
    "worker_shutdown_timeout": {
        "type": "int",
        "description": "Time to wait during shutdown to finish processing requests",
        "default": 5,
    },
    "max_attempts": {
        "type": "int",
        "description": "Number of times to attempt a request update",
        "default": -1,
    },
    "max_timeout": {
        "type": "int",
        "description": "Maximum amount of time to wait between request update retries",
        "default": 30,
    },
    "starting_timeout": {
        "type": "int",
        "description": "Initial amount of time to wait before request update retry",
        "default": 5,
    },
    "working_directory": {
        "type": "str",
        "description": "Working directory to use as a staging area for file parameters",
        "required": False,
    },
}

_MQ_SPEC = {
    "type": "dict",
    "items": {
        "max_attempts": {
            "type": "int",
            "description": "Number of times to attempt reconnection to message queue"
            "before giving up (default -1 aka never)",
            "default": -1,
        },
        "max_timeout": {
            "type": "int",
            "description": "Maximum amount of time to wait between reconnect tries",
            "default": 30,
        },
        "starting_timeout": {
            "type": "int",
            "description": "Initial amount of time to wait before reconnect try",
            "default": 5,
        },
    },
}

SPECIFICATION = {"mq": _MQ_SPEC}

SPECIFICATION.update(_SYSTEM_SPEC)
SPECIFICATION.update(_PLUGIN_SPEC)
SPECIFICATION.update(_CONNECTION_SPEC)



# -*- coding: utf-8 -*-
"""This module currently exists to maintain backwards compatibility."""
import warnings

from brewtils.pika import PikaClient

__all__ = ["PikaClient"]

warnings.warn(
    "This module has been migrated to brewtils.pika and will be removed in a future "
    "release. Please import directly from the new module.",
    DeprecationWarning,
    stacklevel=2,
)



# -*- coding: utf-8 -*-

import collections
import logging
from typing import Any, Dict, List, Mapping

from brewtils.models import Parameter, Resolvable
from brewtils.resolvers.bytes import BytesResolver
from brewtils.resolvers.chunks import ChunksResolver
from brewtils.resolvers.identity import IdentityResolver
from brewtils.schema_parser import SchemaParser


def build_resolver_map(easy_client=None):
    """Builds all resolvers"""

    return [
        IdentityResolver(),  # This should always be first
        BytesResolver(easy_client),
        ChunksResolver(easy_client),
    ]


class ResolutionManager(object):
    """Parameter resolution manager

    This class is used under-the-hood for various plugin functions. Its purpose is to
    remove all the various cleanup and housekeeping steps involved in resolving
    parameters. An example of an unresolved parameter is a dictionary which represents a
    bytes object. In this case the user wants the open file descriptor, not the random
    dictionary that they don't know how to process. The parameter resolver helps handle
    these scenarios.

    This is intended for internal use for the plugin class.
    """

    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.resolvers = build_resolver_map(**kwargs)

    def resolve(self, values, definitions=None, upload=True):
        # type: (Mapping[str, Any], List[Parameter], bool) -> Dict[str, Any]
        """Iterate through parameters, resolving as necessary

        Args:
            values: Dictionary of request parameter values
            definitions: Parameter definitions
            upload: Controls which methods will be called on resolvers

        Returns:
            The resolved parameter dict
        """
        resolved_parameters = {}

        for key, value in values.items():
            # First find the matching Parameter definition, if possible
            definition = Parameter()
            for param_def in definitions or []:
                if param_def.key == key:
                    definition = param_def
                    break

            # Check to see if this is a nested parameter
            if isinstance(value, collections.Mapping) and definition.parameters:
                resolved = self.resolve(
                    value, definitions=definition.parameters, upload=upload
                )

            # See if this is a multi parameter
            elif isinstance(value, list):
                # This is kind of gross because multi-parameters are kind of gross
                # We have to wrap everything into the correct form and pull it out
                resolved = []

                for item in value:
                    resolved_item = self.resolve(
                        {key: item}, definitions=definitions, upload=upload
                    )
                    resolved.append(resolved_item[key])

            # This is a simple parameter
            else:
                # See if this is a parameter that needs to be resolved
                for resolver in self.resolvers:
                    if upload and resolver.should_upload(value, definition):
                        resolvable = resolver.upload(value, definition)
                        resolved = SchemaParser.serialize(resolvable, to_string=False)
                        break
                    elif not upload and resolver.should_download(value, definition):
                        resolvable = Resolvable(**value)
                        resolved = resolver.download(resolvable, definition)
                        break

                # Just a normal parameter
                else:
                    resolved = value

            resolved_parameters[key] = resolved

        return resolved_parameters
